from typing import Tuple, List
from typing_extensions import TypedDict


class ApiException(Exception):
    """Base class for API exceptions. Contains indication of HTTP status.

    Attributes:
        status_code: HTTP status code
    """

    def __init__(self, message: str, status: int):
        """Inits ApiException

        Args:
            message: Exception message.
            status: HTTP status.
        """
        super().__init__(message)
        self.status_code = status

    @property
    def code(self) -> str:
        """Returns exception code used for i18n

        Returns:
            Exception code.
        """
        return self._code

    @code.setter
    def code(self, code: str):
        """Sets error code, used for i18n

        Args:
            code: Error code for i18n.
        """
        self._code = code

    @property
    def arguments(self) -> Tuple:
        """Returns message arguments for i18n

        Returns:
            Message arguments for i18n.
        """
        return self.args

    @arguments.setter
    def arguments(self, args: Tuple):
        """Set message arguments for i18n

        Args:
            args: Arguments for i18n.
        """
        self.args = args


class NotFoundException(ApiException):
    """Throwing this exception results in 404 (Not Found) HTTP response code."""

    def __init__(self, message: str):
        """Inits not found exception.

        Args:
            message: Exception message.
        """
        super().__init__(message, 404)


class ForbiddenException(ApiException):
    """Throwing this exception results in 403 (Forbidden) HTTP response code."""

    def __init__(self, message: str):
        """Inits forbidden exception.

        Args:
            message: Exception message.
        """
        super().__init__(message, 403)


class UnauthorizedException(ApiException):
    """Throwing this exception results in 401 (Unauthorized) HTTP response code."""

    def __init__(self, message: str):
        """Inits unauthorized exception.

        Args:
            message: Exception message.
        """
        super().__init__(message, 401)


class ValidationException(ApiException):
    """Represents validation exception. Throwing this exception results in 400 (Bad Request) HTTP response code.

    Attributes:
        _details: Validation exception details
    """

    def __init__(self, message: str, details: List[dict] = None):
        """Inits validation error.

        Args:
            message: Exception message.
            details: Exception data.
        """
        super().__init__(message + ', check error.details for more information', 400)
        self._details = details

    @property
    def details(self) -> List[dict]:
        """Returns validation exception details.

        Returns:
            Validation exception details.
        """
        return self._details


class InternalException(ApiException):
    """Represents unexpected exception. Throwing this error results in 500 (Internal Error) HTTP response code."""

    def __init__(self, message: str):
        """Inits unexpected exception.

        Args:
            message: Exception message.
        """
        super().__init__(message, 500)


class TooManyRequestsErrorMetadata(TypedDict):
    """Metadata for too many requests error."""
    periodInMinutes: float
    """Throttling period in minutes."""
    requestsPerPeriodAllowed: int
    """Available requests for periodInMinutes."""
    recommendedRetryTime: str
    """Recommended date to retry request."""


class TooManyRequestsException(ApiException):
    """Represents too many requests error. Throwing this exception results in 429 (Too Many Requests) HTTP response
    code."""

    def __init__(self, message: str, metadata: TooManyRequestsErrorMetadata):
        """Inits too many requests exception.

        Args:
            message: Exception message.
            metadata: Error metadata.
        """
        super().__init__(message, 429)
        self.metadata = metadata



import pytest
import respx
from httpx import Response
from mock import MagicMock, AsyncMock
from .httpClient import HttpClient
from .metaStats_client import MetaStatsClient

token = 'header.payload.sign'
expected = {
    'trades': 10,
    'equity': 10102.5,
    'balance': 10105,
    'profit': 104,
    'deposits': 10001
}
account_id = '1234567'
host = 'agiliumtrade.ai'
domain_client = MagicMock()
metastats_client = MetaStatsClient(domain_client)
domain_call_mock = MagicMock()
domain_client.request_metastats = domain_call_mock


@pytest.fixture(autouse=True)
async def run_around_tests():
    global domain_call_mock
    domain_call_mock = AsyncMock(return_value={'metrics': expected})
    global domain_client
    domain_client = MagicMock()
    domain_client.token = token
    domain_client.request_metastats = domain_call_mock
    global metastats_client
    metastats_client = MetaStatsClient(domain_client)
    yield


class TestGetMetrics:
    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_metrics(self):
        """Should retrieve account metrics from API."""
        metrics = await metastats_client.get_metrics(account_id)
        assert metrics == expected
        assert domain_call_mock.call_args[0][0](host, account_id) == {
            'url': 'agiliumtrade.ai/users/current/accounts/1234567/metrics',
            'method': 'GET',
            'headers': {'auth-token': token},
            'params': {'includeOpenPositions': False}
        }
        assert domain_call_mock.call_args[0][1] == '1234567'

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_metrics_with_open_positions(self):
        """Should retrieve account metrics with included open positions from API."""
        expected['inclusive'] = False
        metrics = await metastats_client.get_metrics(account_id, True)
        assert metrics == expected
        assert domain_call_mock.call_args[0][0](host, account_id) == {
            'url': 'agiliumtrade.ai/users/current/accounts/1234567/metrics',
            'method': 'GET',
            'headers': {'auth-token': token},
            'params': {'includeOpenPositions': True}
        }
        assert domain_call_mock.call_args[0][1] == '1234567'


start_time = '2020-01-01 00:00:00.000'
end_time = '2021-01-01 00:00:00.000'


class TestGetTrades:

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_account_trades(self):
        """Should retrieve account trades from API."""
        expected = [{'_id': '1'}]
        domain_call_mock = AsyncMock(return_value={'trades': expected})
        domain_client.request_metastats = domain_call_mock
        trades = await metastats_client.get_account_trades(account_id, start_time, end_time)
        assert trades == expected
        assert domain_call_mock.call_args[0][0](host, account_id) == {
            'url': f'agiliumtrade.ai/users/current/accounts/1234567/historical-trades/{start_time}/{end_time}',
            'method': 'GET',
            'headers': {'auth-token': token},
            'params': {'updateHistory': True, 'limit': 1000, 'offset': 0}
        }
        assert domain_call_mock.call_args[0][1] == '1234567'


class TestGetOpenTrades:

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_account_open_trades(self):
        """Should retrieve account open trades from API."""
        expected = [{'_id': '1'}]
        domain_call_mock = AsyncMock(return_value={'openTrades': expected})
        domain_client.request_metastats = domain_call_mock
        open_trades = await metastats_client.get_account_open_trades(account_id)
        assert open_trades == expected
        assert domain_call_mock.call_args[0][0](host, account_id) == {
            'url': 'agiliumtrade.ai/users/current/accounts/1234567/open-trades',
            'method': 'GET',
            'headers': {'auth-token': token}
        }
        assert domain_call_mock.call_args[0][1] == '1234567'



from datetime import datetime
from typing import List
from typing_extensions import TypedDict
from .errorHandler import ValidationException


class AccountInfo(TypedDict):
    """Account request info."""
    id: str
    """Primary account id."""
    regions: List[str]
    """Account available regions."""


class DomainClient:
    """Connection URL and request managing client"""

    def __init__(self, http_client, token: str, domain: str = None):
        """Inits domain client instance.

        Args:
            http_client: HTTP client.
            token: Authorization token.
            domain: Domain to connect to, default is agiliumtrade.agiliumtrade.ai.
        """
        self._httpClient = http_client
        self._domain = domain or 'agiliumtrade.agiliumtrade.ai'
        self._token = token
        self._urlCache = None
        self._accountCache = {}

    @property
    def token(self) -> str:
        """Returns domain client token.

        Returns:
            Client token.
        """
        return self._token

    async def request_metastats(self, get_opts, account_id: str):
        """Sends a MetaStats API request.

        Args:
            get_opts: Function to get request options.
            account_id: Account id..

        Returns:
            Request result.
        """
        await self._update_host()
        await self._update_account_host(account_id)
        account_cache = self._accountCache[account_id]
        try:
            region_settings = account_cache['regions'][account_cache['regionIndex']]
            opts = get_opts(f'https://metastats-api-v1.{region_settings["region"]}.{self._urlCache["domain"]}',
                            region_settings['id'])

            return await self._httpClient.request(opts)
        except Exception as err:
            if err.__class__.__name__ not in ['ConflictException', 'InternalException', 'ApiException',
                                              'ConnectTimeout']:
                raise err
            else:
                if len(account_cache['regions']) == account_cache['regionIndex'] + 1:
                    raise err
                else:
                    account_cache['regionIndex'] += 1
                    return await self.request_metastats(get_opts, account_id)

    async def _update_host(self):
        if not self._urlCache or self._urlCache['lastUpdated'] < datetime.now().timestamp() - 60 * 10:
            url_settings = await self._httpClient.request_with_failover({
                'url': f'https://mt-provisioning-api-v1.{self._domain}/users/current/servers/mt-client-api',
                'method': 'GET',
                'headers': {
                    'auth-token': self._token
                }
            })
            self._urlCache = {
                'domain': url_settings['domain'],
                'lastUpdated': datetime.now().timestamp()
            }

    async def _update_account_host(self, account_id: str):
        if account_id not in self._accountCache or self._accountCache[account_id]['lastUpdated'] < \
                datetime.now().timestamp() - 60 * 10:

            async def get_account(id: str):
                account_opts = {
                    'url': f'https://mt-provisioning-api-v1.{self._domain}/users/current/accounts/{id}',
                    'method': 'GET',
                    'headers': {
                        'auth-token': self._token
                    },
                }

                return await self._httpClient.request_with_failover(account_opts)

            account_data = await get_account(account_id)
            if 'primaryAccountId' in account_data:
                account_data = await get_account(account_data['primaryAccountId'])

            accounts = [{'_id': account_data['_id'], 'region': account_data['region'],
                         'state': account_data['state']}] + (
                account_data['accountReplicas'] if 'accountReplicas' in account_data else [])
            accounts = list(filter(lambda account: account['state'] == 'DEPLOYED', accounts))

            if not len(accounts):
                raise ValidationException('There are no replicas deployed yet. Please make sure at least ' +
                                          'one of the replicas is deployed.')

            regions = list(map(lambda account: {'region': account['region'], 'id': account['_id']}, accounts))

            self._accountCache[account_id] = {
                'regions': regions,
                'regionIndex': 0,
                'lastUpdated': datetime.now().timestamp()
            }



from metaapi_cloud_sdk import MetaStats, MetaApi
import asyncio
import os


async def example():
    # your MetaApi API token
    token = os.getenv('TOKEN') or '<put in your token here>'
    # your MetaApi account id
    account_id = os.getenv('ACCOUNT_ID') or '<put in your account id here>'
    # your MetaApi account
    account = None

    api = MetaApi(token)
    meta_stats = MetaStats(token)
    # you can configure http client via second parameter,
    # see in-code documentation for full definition of possible configuration options

    async def account_deploy():
        try:
            nonlocal account
            account = await api.metatrader_account_api.get_account(account_id)

            #  wait until account is deployed and connected to broker
            print('Deploying account')
            if account.state != 'DEPLOYED':
                await account.deploy()
            else:
                print('Account already deployed')
            print('Waiting for API server to connect to broker (may take couple of minutes)')
            if account.connection_status != 'CONNECTED':
                await account.wait_connected()

        except Exception as err:
            print(meta_stats.format_error(err))

    await account_deploy()

    async def get_account_metrics():
        try:
            metrics = await meta_stats.get_metrics(account_id)
            print(metrics)  # -> {'trades': ..., 'balance': ..., ...}
        except Exception as err:
            print(meta_stats.format_error(err))

    await get_account_metrics()

    async def get_account_trades(start_time, end_time):
        try:
            trades = await meta_stats.get_account_trades(account_id, start_time, end_time)
            print(trades[-5:])  # -> {_id: ..., gain: ..., ...}
        except Exception as err:
            print(meta_stats.format_error(err))

    await get_account_trades('2020-01-01 00:00:00.000', '2021-01-01 00:00:00.000')

    async def get_account_open_trades():
        try:
            trades = await meta_stats.get_account_open_trades(account_id)
            print(trades)  # -> {_id: ..., gain: ..., ...}
        except Exception as err:
            print(meta_stats.format_error(err))

    await get_account_open_trades()

asyncio.run(example())



from typing_extensions import TypedDict
from typing import Optional, List


class CurrencySummaryHistoryDayMetrics(TypedDict):
    """Profit from trading a currency pair in one trading day."""
    date: str
    """Date of trading day, in broker timezone, YYYY-MM-DD format."""
    totalProfit: float
    """Total profit at the end of the day."""
    totalPips: Optional[float]
    """Total pips of trading day."""
    shortProfit: Optional[float]
    """Total profit of short trades per day."""
    longProfit: Optional[float]
    """Total profit of long trades per day."""
    shortPips: Optional[float]
    """Total pips of short trades per day."""
    longPips: Optional[float]
    """Total pips of long trades per day."""


class CurrencySummaryTotalMetrics(TypedDict):
    """Provides general data of this currency trading."""
    profit: float
    """Cumulative profit of this currency trading."""
    trades: int
    """The number of all trades with this currency."""
    pips: Optional[float]
    """Cumulative pips of this currency trading."""
    wonTrades: Optional[int]
    """The number of winning trades with this currency."""
    lostTrades: Optional[int]
    """The number of losing trades with this currency."""
    wonTradesPercent: Optional[float]
    """Percentage of winning trades with this currency."""
    lostTradesPercent: Optional[float]
    """Percentage of losing trades with this currency."""


class CurrencySummaryTradeMetrics(TypedDict):
    """Provides profit and number of trades in specific trade and currency."""
    profit: float
    """Cumulative profit of this currency trading."""
    trades: int
    """The number of all trades with this currency."""
    pips: Optional[float]
    """Cumulative pips of this currency trading."""


class CurrencySummaryMetrics(TypedDict):
    """Provides statistics on winning and losing trades indicating the amount in the context of long and
    short positions. Statistics is given for all currency pairs, for which positions were opened."""
    currency: str
    """Trading currency pair."""
    history: List[CurrencySummaryHistoryDayMetrics]
    """History of trading a currency pair per trading days."""
    total: CurrencySummaryTotalMetrics
    """General data (such as profit, number of trades) about trading a specific currency pair."""
    short: Optional[CurrencySummaryTradeMetrics]
    """Profit and number of trades of short trades in a specific currency."""
    long: Optional[CurrencySummaryTradeMetrics]
    """Profit and number of trades of long trades in a specific currency."""


class PeriodMetrics(TypedDict):
    """Provides statistics for one trade period compared to the results for the previous period."""
    profit: Optional[float]
    """Cumulative profit of this period."""
    pips: Optional[float]
    """Cumulative pips of this period."""
    lots: Optional[float]
    """Cumulative lots of this period."""
    gain: Optional[float]
    """Gain of this period."""
    trades: Optional[int]
    """The number of trades of this period."""
    wonTradesPercent: Optional[float]
    """Percentage of winning trades of this period."""
    profitDifference: Optional[float]
    """Difference in profit with the previous period."""
    pipsDifference: Optional[float]
    """Difference in pips with the previous period."""
    lotsDifference: Optional[float]
    """Difference in lots with the previous period."""
    gainDifference: Optional[float]
    """Difference in gain with the previous period."""
    tradesDifference: Optional[int]
    """Difference in the number of trades with the previous period."""
    wonTradesPercentDifference: Optional[float]
    """Difference in percentage of winning trades with the previous period."""


class Periods(TypedDict):
    """Provides statistics for today, this week, this month, this year."""
    today: Optional[PeriodMetrics]
    """Trade information for today."""
    thisWeek: Optional[PeriodMetrics]
    """Trade information for this week."""
    thisMonth: Optional[PeriodMetrics]
    """Trade information for this month."""
    thisYear: Optional[PeriodMetrics]
    """Trade information for this year."""


class DailyGrowthMetrics(TypedDict):
    """Provides each profit received from the volume of trade and changes in balance, total accumulated income and
    existing account drawdown by day."""
    date: str
    """Date of trading day in broker timezone, YYYY-MM-DD format."""
    profit: Optional[float]
    """Cumulative profit per day."""
    pips: Optional[float]
    """Cumulative pips per day."""
    lots: Optional[float]
    """Cumulative lots per day."""
    gains: Optional[float]
    """Cumulative gains per day."""
    totalProfit: float
    """Total profit in this day end."""
    totalGains: float
    """Total gains in this day end."""
    balance: float
    """Balance in this day end."""
    drawdownPercentage: Optional[float]
    """Percentage of balance drawdown in this day end."""
    drawdownProfit: Optional[float]
    """Maximum registered balance drawdown in basic currency during this day."""


class MonthlyAnalyticCurrencyMetrics(TypedDict):
    """Currency pair trading information for monthly analysis."""
    currency: str
    """Currency pair."""
    averageHoldingTimeLongsInMilliseconds: Optional[float]
    """Average holding time of long trades."""
    averageHoldingTimeShortsInMilliseconds: Optional[float]
    """Average holding time of short trades."""
    rewardToRiskRatio: float
    """The difference between reward and risk, where the lesser is always one. So 0 means
    reward:risk=1:1, 2 means 3:1, -0.5 means 1:1.5."""
    popularityPercent: float
    """The percentage of popularity of this currency this month."""


class MonthlyAnalyticsMetrics(TypedDict):
    """Monthly analysis of trading on this account."""
    date: str
    """Date of trading month in broker timezone, YYYY-MM format."""
    profit: Optional[float]
    """Cumulative profit per month."""
    pips: Optional[float]
    """Cumulative pips per month."""
    lots: Optional[float]
    """Cumulative lots per month."""
    gains: Optional[float]
    """Cumulative gains per month."""
    trades: Optional[int]
    """The number of trades of this month."""
    currencies: Optional[List[MonthlyAnalyticCurrencyMetrics]]
    """List of currency pair trading informations for monthly analysis."""


class TradeByTimeMetrics(TypedDict):
    """Opening/closing deals by days of the week or by hours of the day."""
    date: str
    """Date of trading month in broker timezone, YYYY-MM format."""
    profit: float
    """The total profit of the trades at this time."""
    shortProfit: Optional[float]
    """The total profit of short trades at this time."""
    longProfit: Optional[float]
    """The total profit of long trades at this time."""
    wonProfit: Optional[float]
    """The total profit of winning trades at this time."""
    lostProfit: Optional[float]
    """The total profit of losing trades at this time."""
    pips: Optional[float]
    """The total pips of the trades at this time."""
    shortPips: Optional[float]
    """The total pips of short trades at this time."""
    longPips: Optional[float]
    """The total pips of long trades at this time."""
    wonPips: Optional[float]
    """The total pips of winning trades at this time."""
    lostPips: Optional[float]
    """The total pips of losing trades at this time."""
    lots: float
    """Cumulative lots of trades at this time."""
    gains: float
    """Cumulative gains of trades at this time."""
    shortGains: Optional[float]
    """Cumulative gains of short trades at this time."""
    longGains: Optional[float]
    """Cumulative gains of long trades at this time."""
    wonGains: Optional[float]
    """Cumulative gains of winning trades at this time."""
    lostGains: Optional[float]
    """Cumulative gains of losing trades at this time."""
    trades: int
    """The number of all trades at this time."""
    shortTrades: Optional[int]
    """The number of short trades at this time."""
    longTrades: Optional[int]
    """The number of long trades at this time."""
    wonTrades: Optional[int]
    """The number of winning trades at this time."""
    lostTrades: Optional[int]
    """The number of losing trades at this time."""
    shortTradesPercent: Optional[float]
    """Percentage of short trades at this time."""
    longTradesPercent: Optional[float]
    """Percentage of long trades at this time."""
    wonTradesPercent: Optional[float]
    """Percentage of winning trades at this time."""
    lostTradesPercent: Optional[float]
    """Percentage of losing trades at this time."""
    hour: Optional[int]
    """Day hour (only for by hour case)', within 0-23."""
    day: Optional[int]
    """Weekday number (only for by day case), within 0-6."""


class RiskOfRuinMetrics(TypedDict):
    """Risk of ruin of balance metrics."""
    lossSize: float
    """Loss size of balance."""
    probabilityOfLoss: float
    """Probability of loss shows the risk of losing a particular part of the balance."""
    consecutiveLosingTrades: float
    """The number of losing trades that must be entered sequentially in order for this part of the
    balance to be lost."""


class OneTradeDurationMetrics(TypedDict):
    """Metrics of one trade duration."""
    gains: List[float]
    """List of gains for this duration."""
    profits: List[float]
    """List of profits for this duration."""
    lots: List[float]
    """List of lots for this duration."""
    pips: Optional[List[float]]
    """List of pips for this duration."""
    durationInMinutes: float
    """Duration of trades in minutes."""


class TradeDurationMetrics(TypedDict):
    """Metrics for each duration of trades."""
    won: Optional[List[OneTradeDurationMetrics]]
    """Metrics of winning trades."""
    lost: Optional[List[OneTradeDurationMetrics]]
    """Metrics of losing trades."""


class TradeDurationDiagramColumnCollectionMetrics(TypedDict):
    """Collection of metrics of trades in the current column for the diagram."""
    gains: List[float]
    """List of gains."""
    profits: List[float]
    """List of profits."""
    lots: List[float]
    """List of lots."""
    pips: Optional[List[float]]
    """List of pips."""


class TradeDurationDiagramColumnMetrics(TypedDict):
    """Information column about the duration of trades for the diagram."""
    durations: float
    """The number of durations in this column."""
    trades: int
    """The number of trades in this column."""
    name: str
    """Name of this column, one of 'seconds', 'minutes', 'hours', 'days', 'weeks', 'months'."""
    minDurationInSeconds: float
    """Minimum trade duration in this column in seconds."""
    maxDurationInSeconds: Optional[float]
    """Maximum trade duration in this column in seconds."""
    won: Optional[TradeDurationDiagramColumnCollectionMetrics]
    """Collection of metrics of winning trades in this column."""
    lost: Optional[TradeDurationDiagramColumnCollectionMetrics]
    """Collection of metrics of losing trades in this column."""


class Metrics(TypedDict):
    """Trading statistics metrics."""
    inclusive: Optional[bool]
    """Indicates whether open positions are included in the metrics, "False" means that there are no open positions.
    Only for a request with includeOpenPositions=True."""
    balance: float
    """Money on the account, not accounting for the results of currently open positions."""
    highestBalanceDate: Optional[str]
    """Date of maximum balance that have ever been on the account, in broker timezone, YYYY-MM-DD HH:mm:ss.SSS
    format."""
    highestBalance: Optional[float]
    """Maximum balance that have ever been on the account."""
    equity: float
    """The result (current amount) of all positions, including opened."""
    margin: float
    """Current value of margin."""
    freeMargin: float
    """Current value of free margin."""
    marginLevel: Optional[float]
    """Current value of margin level."""
    trades: float
    """Total number of closed positions on the account."""
    withdrawals: Optional[float]
    """Total amount withdrawn from the deposit."""
    averageTradeLengthInMilliseconds: Optional[float]
    """Average trade length (time from open to close) in milliseconds."""
    bestTrade: Optional[float]
    """The best profit from one trade that has ever been on the account."""
    worstTrade: Optional[float]
    """The worst profit from one trade that has ever been on the account."""
    bestTradePips: Optional[float]
    """The best pips from one trade that has ever been on the account."""
    worstTradePips: Optional[float]
    """The worst pips from one trade that has ever been on the account."""
    bestTradeDate: Optional[str]
    """Date of the best profit from one trade that have ever been on the account, in broker timezone,
    YYYY-MM-DD HH:mm:ss.SSS format."""
    bestTradePipsDate: Optional[str]
    """Date of the best pips from one trade that have ever been on the account, in broker timezone,
    YYYY-MM-DD HH:mm:ss.SSS format."""
    worstTradeDate: Optional[str]
    """Date of the worst profit from one trade that have ever been on the account, in broker timezone,
    YYYY-MM-DD HH:mm:ss.SSS format."""
    worstTradePipsDate: Optional[str]
    """Date of the worst pips from one trade that have ever been on the account, in broker timezone,
    YYYY-MM-DD HH:mm:ss.SSS format."""
    cagr: Optional[float]
    """Compound annual growth rate."""
    commissions: Optional[float]
    """Commissions charged by the broker for the entire period."""
    dailyGain: Optional[float]
    """Compound daily rate of return."""
    monthlyGain: Optional[float]
    """Compound monthly rate of return."""
    equityPercent: Optional[float]
    """Percentage of current equity to balance."""
    expectancy: Optional[float]
    """The average expected profitability of one trade in basic currency."""
    expectancyPips: Optional[float]
    """The average expected profitability of one trade in pips."""
    gain: Optional[float]
    """Time-weighted rate of return."""
    geometricHoldingPeriodReturn: Optional[float]
    """Geometric holding period return."""
    interest: Optional[float]
    """Cumulative interest and swap for the entire period."""
    longTrades: Optional[int]
    """The number of long trades."""
    shortTrades: Optional[int]
    """The number of long trades."""
    longWonTrades: Optional[int]
    """The number of long winning trades."""
    shortWonTrades: Optional[int]
    """The number of short winning trades."""
    longWonTradesPercent: Optional[float]
    """Percentage of long winning trades."""
    shortWonTradesPercent: Optional[float]
    """Percentage of short winning trades."""
    maxDrawdown: Optional[float]
    """Percentage of maximum drawdown of balance during the entire trading history."""
    mar: Optional[float]
    """Mar ratio."""
    lots: Optional[float]
    """Total volume of trades."""
    pips: Optional[float]
    """Cumulative price units."""
    profit: float
    """The total yield of closed positions for the entire period (total result)."""
    deposits: float
    """Cumulative deposit for the entire period."""
    absoluteGain: Optional[float]
    """Simple deposit increase without regard to reinvestment."""
    profitFactor: Optional[float]
    """The amount yielded by winning trades divided by the amount of losses yielded by losing trades. Result in range
    0 - Infinity means: `0` - only loss, `1` - profit equals to loss, `Infinity` - only profit."""
    sharpeRatio: Optional[float]
    """Average return earned in excess of the risk-free rate per unit of volatility. It is calculated if there are at
    least 30 closed deals in the history."""
    sortinoRatio: Optional[float]
    """Differentiates harmful volatility from total overall volatility. It is calculated if there are at least 30
    closed deals in the history."""
    standardDeviationProfit: Optional[float]
    """Statistical measure of volatility shows how much variation or dispersion. It is calculated if there are at
    least 30 closed deals in the history."""
    kurtosisProfit: Optional[float]
    """A statistical measure that is used to describe profit distribution. It is calculated if there are at least
    30 closed deals in the history."""
    averageHoldingPeriodReturn: Optional[float]
    """Average holding period return. It is calculated if there are at least 30 closed deals in the history."""
    averageWin: Optional[float]
    """Average win in basic currency."""
    averageWinPips: Optional[float]
    """Average win in pips."""
    averageLoss: Optional[float]
    """Average loss in basic currency."""
    averageLossPips: Optional[float]
    """Average loss in pips."""
    wonTradesPercent: Optional[float]
    """Percentage of winning trades."""
    lostTradesPercent: Optional[float]
    """Percentage of losing trades."""
    zScore: Optional[float]
    """Ability of a trading system to generate wins and losses in streaks. It is calculated if there are at least 30
    closed deals in the history."""
    probability: Optional[float]
    """Probability that a profit will be followed by a profit and a loss by a loss."""
    daysSinceTradingStarted: Optional[float]
    """The number of days that have passed since the opening of the first trade"""
    currencySummary: Optional[List[CurrencySummaryMetrics]]
    """Currency trading summary."""
    dailyGrowth: Optional[List[DailyGrowthMetrics]]
    """Daily gain shows the change in account profitability on trading days."""
    monthlyAnalytics: Optional[List[MonthlyAnalyticsMetrics]]
    """Monthly analysis of trading on this account."""
    closeTradesByWeekDay: Optional[List[TradeByTimeMetrics]]
    """Closing deals by days of the week."""
    openTradesByHour: Optional[List[TradeByTimeMetrics]]
    """Opening deals by hour of the day."""
    periods: Optional[Periods]
    """Trading stats for a few periods compared to the results for the previous period."""
    riskOfRuin: Optional[List[RiskOfRuinMetrics]]
    """Risk of ruin of balance."""
    tradeDuration: Optional[TradeDurationMetrics]
    """Metrics for each duration of trades."""
    tradeDurationDiagram: Optional[List[TradeDurationDiagramColumnMetrics]]
    """List of information columns about the duration of trades for the diagram."""


class Trade(TypedDict):
    """Historical trade."""
    _id: str
    """Historical trade id."""
    accountId: str
    """MetaApi account id."""
    volume: float
    """Trade volume."""
    durationInMinutes: float
    """Trade duration in minutes."""
    profit: float
    """Trade profit."""
    gain: float
    """Trade gain."""
    success: str
    """Trade success."""
    openTime: str
    """Time the trade was opened at in broker timezone, YYYY-MM-DD HH:mm:ss.SSS format."""
    type: str
    """Trade type"""
    symbol: Optional[str]
    """Symbol the trade relates to."""
    closeTime: Optional[str]
    """Time the trade was closed at in broker timezone, YYYY-MM-DD HH:mm:ss.SSS format."""
    openPrice: Optional[float]
    """Trade opening price."""
    closePrice: Optional[float]
    """Trade closing price."""
    pips: Optional[float]
    """The number of pips earned (positive) or lost (negative) in this trade."""
    riskInBalancePercent: Optional[float]
    """Trade risk in % of balance."""
    riskInPips: Optional[float]
    """Trade risk in pips."""


class OpenTrade(TypedDict):
    """Open trade."""
    _id: str
    """Historical trade id."""
    accountId: str
    """MetaApi account id."""
    volume: float
    """Trade volume."""
    durationInMinutes: float
    """Trade duration in minutes."""
    profit: float
    """Trade profit."""
    gain: float
    """Trade gain."""
    success: str
    """Trade success."""
    openTime: str
    """Time the trade was opened at in broker timezone, YYYY-MM-DD HH:mm:ss.SSS format."""
    type: str
    """Trade type"""
    symbol: Optional[str]
    """Symbol the trade relates to."""
    openPrice: Optional[float]
    """Trade opening price."""
    pips: Optional[float]
    """The number of pips earned (positive) or lost (negative) in this trade."""
    riskInBalancePercent: Optional[float]
    """Trade risk in % of balance."""
    riskInPips: Optional[float]
    """Trade risk in pips."""



from .httpClient import HttpClient
from .domain_client import DomainClient
from mock import AsyncMock, ANY
import pytest
from freezegun import freeze_time
import respx
import json
from httpx import Response
import asyncio

http_client = HttpClient()
domain_client = DomainClient(http_client, 'token')
request_mock = AsyncMock()

token = 'header.payload.sign'
expected = {'trades': 10, 'equity': 10102.5, 'balance': 10105, 'profit': 104, 'deposits': 10001}
start_time = '2020-10-05 10:00:00.000'

request_call: respx.Route
host_call: respx.Route
get_account_call: respx.Route


def get_opts(host, id):
    return {
        'url': host + f'/users/current/accounts/{id}/open-trades',
        'method': 'GET',
        'headers': {
          'auth-token': token
        },
    }


@pytest.fixture(autouse=True)
async def run_around_tests():
    global http_client
    http_client = HttpClient()
    global domain_client
    domain_client = DomainClient(http_client, token)

    global get_account_call
    get_account_call = respx.get('https://mt-provisioning-api-v1.agiliumtrade.agiliumtrade.ai/users/current/'
                                 'accounts/accountId')\
        .mock(return_value=Response(200, content=json.dumps({
            '_id': 'accountId', 'region': 'vint-hill', 'state': 'DEPLOYED', 'accountReplicas': [
                {'_id': 'accountId2', 'region': 'us-west', 'state': 'DEPLOYED'}]})))

    global request_call
    request_call = respx.get('https://metastats-api-v1.vint-hill.agiliumtrade.ai/users/current/accounts/accountId/'
                             'open-trades').mock(return_value=Response(200, content=json.dumps(expected)))

    global host_call
    host_call = respx.get(
        'https://mt-provisioning-api-v1.agiliumtrade.agiliumtrade.ai/users/current/servers/mt-client-api')\
        .mock(return_value=Response(200, content=json.dumps({'domain': 'agiliumtrade.ai'})))
    yield


class TestDomainClient:

    @respx.mock
    @pytest.mark.asyncio
    async def test_execute_request(self):
        """Should execute request."""
        response = await domain_client.request_metastats(get_opts, 'accountId')
        assert response == expected
        assert request_call.call_count == 1

    @respx.mock
    @pytest.mark.asyncio
    async def test_request_url_again_if_expired(self):
        """Should request url again if expired."""
        with freeze_time(start_time) as frozen_datetime:
            await domain_client.request_metastats(get_opts, 'accountId')
            frozen_datetime.tick(610)
            response = await domain_client.request_metastats(get_opts, 'accountId')
            assert response == expected
            assert host_call.call_count == 2

    @respx.mock
    @pytest.mark.asyncio
    async def test_use_cached_url(self):
        """Should use cached url on repeated request."""
        await domain_client.request_metastats(get_opts, 'accountId')
        response = await domain_client.request_metastats(get_opts, 'accountId')
        assert response == expected
        assert host_call.call_count == 1

    @respx.mock
    @pytest.mark.asyncio
    async def test_return_error_if_failed_host(self):
        """Should return error if failed to get host."""
        host_call.mock(return_value=Response(400))
        try:
            await domain_client.request_metastats(get_opts, 'accountId')
            pytest.fail()
        except Exception as err:
            assert err.__class__.__name__ == 'ValidationException'

    @respx.mock
    @pytest.mark.asyncio
    async def test_return_error_if_failed_get_account_data(self):
        """Should return error if failed to get account data."""
        get_account_call.mock(return_value=Response(400))
        try:
            await domain_client.request_metastats(get_opts, 'accountId')
            pytest.fail()
        except Exception as err:
            assert err.__class__.__name__ == 'ValidationException'

    @respx.mock
    @pytest.mark.asyncio
    async def test_request_main_account_if_using_replica(self):
        """Should request main account if using replica."""
        respx.get(
            'https://mt-provisioning-api-v1.agiliumtrade.agiliumtrade.ai/users/current/accounts/accountId2') \
            .mock(return_value=Response(200, content=json.dumps({
                '_id': 'accountId2', 'region': 'us-west', 'primaryAccountId': 'accountId', 'state': 'DEPLOYED'})))

        response = await domain_client.request_metastats(get_opts, 'accountId2')
        assert response == expected

    @respx.mock
    @pytest.mark.asyncio
    async def test_try_another_region_if_first_failed(self):
        """Should try another region if the first failed."""
        request_call.mock(return_value=Response(500))
        respx.get(
            'https://metastats-api-v1.us-west.agiliumtrade.ai/users/current/accounts/accountId2/open-trades') \
            .mock(return_value=Response(200, content=json.dumps(expected)))
        response = await domain_client.request_metastats(get_opts, 'accountId')
        assert response == expected
        assert host_call.call_count == 1

    @respx.mock
    @pytest.mark.asyncio
    async def test_return_error_if_all_regions_failed(self):
        """Should return error if all regions failed."""
        request_call.mock(return_value=Response(500))
        respx.get(
            'https://metastats-api-v1.us-west.agiliumtrade.ai/users/current/accounts/accountId2/open-trades') \
            .mock(return_value=Response(500))
        try:
            await domain_client.request_metastats(get_opts, 'accountId')
            pytest.fail()
        except Exception as err:
            assert err.__class__.__name__ == 'InternalException'

    @respx.mock
    @pytest.mark.asyncio
    async def test_update_host_if_expired(self):
        """Should execute a request and update host if expired."""
        with freeze_time(start_time) as frozen_datetime:
            other_region_call = respx.get('https://metastats-api-v1.us-west.agiliumtrade.ai/' +
                                          'users/current/accounts/accountId2/open-trades')\
                .mock(return_value=Response(200, content=json.dumps(expected)))
            get_account_call.mock(return_value=Response(200, content=json.dumps({
                '_id': 'accountId', 'region': 'vint-hill', 'state': 'DEPLOYED'})))

            await domain_client.request_metastats(get_opts, 'accountId')
            await asyncio.sleep(0.05)
            assert get_account_call.call_count == 1
            assert other_region_call.call_count == 0
            get_account_call.mock(return_value=Response(200, content=json.dumps({
                '_id': 'accountId2', 'region': 'us-west', 'state': 'DEPLOYED'})))
            frozen_datetime.tick(610)
            result = await domain_client.request_metastats(get_opts, 'accountId')
            assert result == expected
            await asyncio.sleep(0.05)
            assert get_account_call.call_count == 2
            assert other_region_call.call_count == 1
            await domain_client.request_metastats(get_opts, 'accountId')
            await asyncio.sleep(0.05)
            assert get_account_call.call_count == 2
            assert other_region_call.call_count == 2

    @respx.mock
    @pytest.mark.asyncio
    async def test_return_error_if_no_replicas_deployed(self):
        """Should return error if no replicas are deployed."""
        get_account_call.mock(return_value=Response(200, content=json.dumps({
            '_id': 'accountId', 'region': 'vint-hill', 'state': 'UNDEPLOYED'})))
        try:
            await domain_client.request_metastats(get_opts, 'accountId')
            pytest.fail()
        except Exception as err:
            assert err.__class__.__name__ == 'ValidationException'

    @respx.mock
    @pytest.mark.asyncio
    async def test_filter_out_undeployed_replicas(self):
        """Should filter out undeployed replicas."""
        get_account_call.mock(return_value=Response(200, content=json.dumps({
            '_id': 'accountId', 'region': 'vint-hill', 'state': 'UNDEPLOYED', 'accountReplicas': [
                {'_id': 'accountId2', 'region': 'us-west', 'state': 'UNDEPLOYED'},
                {'_id': 'accountId3', 'region': 'germany', 'state': 'DEPLOYED'}
            ]})))
        other_region_call = respx\
            .get('https://metastats-api-v1.germany.agiliumtrade.ai/users/current/accounts/accountId3/open-trades')\
            .mock(return_value=Response(200, content=json.dumps(expected)))
        response = await domain_client.request_metastats(get_opts, 'accountId')
        assert response == expected
        assert other_region_call.call_count == 1



from .clients.httpClient import HttpClient, RetryOpts
from .clients.metaStats_client import MetaStatsClient
from .clients.domain_client import DomainClient
from typing_extensions import TypedDict
from typing import Optional
import traceback


class ConnectionOptions(TypedDict):
    """Connection options."""
    requestTimeout: Optional[float]
    """Request timeout in seconds, default is 60."""
    domain: Optional[str]
    """Request domain, default 'agiliumtrade.agiliumtrade.ai.'"""
    retryOpts: Optional[RetryOpts]
    """Retry options."""


class MetaStats:
    """MetaStats API SDK."""

    def __init__(self, token: str, opts: ConnectionOptions = None):
        """Inits MetaStats class instance.

        Args:
            token: Authorization token.
            opts: Connection options.
        """
        opts: ConnectionOptions = opts or {}
        http_client = HttpClient(opts['requestTimeout'] if 'requestTimeout' in opts else None,
                                 opts['retryOpts'] if 'retryOpts' in opts else None)
        domain_client = DomainClient(http_client, token, opts['domain'] if 'domain' in opts else None)
        self._metaStatsClient = MetaStatsClient(domain_client)

    @property
    def get_metrics(self):
        """Returns the get_metrics MetaStatsClient method bound to the MetaStatsClient instance.

        Returns:
            get_metrics MetaStatsClient method.
        """
        return self._metaStatsClient.get_metrics

    @property
    def get_account_trades(self):
        """Returns the get_account_trades MetaStatsClient method bound to the MetaStatsClient instance.

        Returns:
            get_account_trades MetaStatsClient method.
        """
        return self._metaStatsClient.get_account_trades

    @property
    def get_account_open_trades(self):
        """Returns the get_account_open_trades MetaStatsClient method bound to the MetaStatsClient instance.

        Returns:
            get_account_open_trades MetaStatsClient method.
        """
        return self._metaStatsClient.get_account_open_trades

    @staticmethod
    def format_error(err: Exception):
        """Formats and outputs metaApi errors with additional information.

        Args:
            err: Exception to process.
        """
        error = {'name': err.__class__.__name__, 'message': err.args[0]}
        if hasattr(err, 'status_code'):
            error['status_code'] = err.status_code
        if err.__class__.__name__ == 'ValidationException':
            error['details'] = err.details
        if err.__class__.__name__ == 'TooManyRequestsException':
            error['metadata'] = err.metadata
        error['trace'] = traceback.format_exc()
        return error



from datetime import datetime
from typing_extensions import TypedDict
from typing import List, Optional
import iso8601
import pytz


def date(date_time: str or float or int) -> datetime:
    """Parses a date string into a datetime object."""
    if isinstance(date_time, float) or isinstance(date_time, int):
        return datetime.fromtimestamp(max(date_time, 100000)).astimezone(pytz.utc)
    else:
        return iso8601.parse_date(date_time)


def format_date(date: datetime) -> str:
    """Converts date to format compatible with JS"""
    return date.astimezone(pytz.utc).isoformat(timespec='milliseconds').replace('+00:00', 'Z')


class ValidationDetails(TypedDict):
    """Object to supply additional information for validation exceptions."""
    parameter: str
    """Name of invalid parameter."""
    value: Optional[str]
    """Entered invalid value."""
    message: str
    """Error message."""


class ExceptionMessage(TypedDict):
    """A REST API response that contains an exception message"""
    id: int
    """Error id"""
    error: str
    """Error name"""
    numericCode: Optional[int]
    """Numeric error code"""
    stringCode: Optional[str]
    """String error code"""
    message: str
    """Human-readable error message"""
    details: Optional[List[ValidationDetails]]
    """Additional information about error. Used to supply validation error details."""



from metaapi_cloud_sdk import MetaStats, MetaApi
import asyncio
import os


async def example_request():
    # your MetaApi API token
    token = os.getenv('TOKEN') or '<put in your token here>'
    # your MetaApi account id
    account_id = os.getenv('ACCOUNT_ID') or '<put in your account id here>'

    api = MetaApi(token)
    meta_stats = MetaStats(token)
    # you can configure http client via second parameter,
    # see in-code documentation for full definition of possible configuration options

    try:
        account = await api.metatrader_account_api.get_account(account_id)

        #  wait until account is deployed and connected to broker
        print('Deploying account')
        if account.state != 'DEPLOYED':
            await account.deploy()
        else:
            print('Account already deployed')
        print('Waiting for API server to connect to broker (may take couple of minutes)')
        if account.connection_status != 'CONNECTED':
            await account.wait_connected()

        metrics = await meta_stats.get_metrics(account_id)
        print(metrics)  # -> {'trades': ..., 'balance': ..., ...}

        trades = await meta_stats.get_account_trades(account_id, '0000-01-01 00:00:00.000', '9999-01-01 00:00:00.000')
        print(trades[-5:])  # -> {_id: ..., gain: ..., ...}

        open_trades = await meta_stats.get_account_open_trades(account_id)
        print(open_trades)  # -> {_id: ..., gain: ..., ...}

    except Exception as err:
        print(meta_stats.format_error(err))
    exit(0)

asyncio.run(example_request())



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

install_requires = ['typing-extensions>=4,<5', 'iso8601', 'pytz', 'httpx==0.23.0']

tests_require = ['pytest==6.2.5', 'pytest-mock==3.8.2', 'pytest-asyncio==0.16.0', 'mock==4.0.3', 'respx==0.19.2',
                 'freezegun==1.0.0']

setuptools.setup(
    name="metaapi_cloud_metastats_sdk",
    version="3.3.0",
    author="MetaApi DMCC",
    author_email="support@metaapi.cloud",
    description="Python SDK for MetaStats forex trading statistics API. Can calculate metrics for MetaTrader "
                "accounts added to MetaApi. Supports both MetaTrader 5 (MT5) and MetaTrader 4 (MT4). "
                "(https://metaapi.cloud)",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    keywords=['metaapi.cloud', 'MetaTrader', 'MetaTrader 5', 'MetaTrader 4', 'MetaTrader5', 'MetaTrader4', 'MT', 'MT4',
              'MT5', 'forex', 'API', 'REST', 'client', 'sdk', 'cloud', 'metrics', 'MetaStats', 'metastats'],
    url="https://github.com/agiliumtrade-ai/metaapi-metastats-python-sdk",
    include_package_data=True,
    package_dir={'metaapi_cloud_metastats_sdk': 'lib'},
    packages=['metaapi_cloud_metastats_sdk'],
    install_requires=install_requires,
    tests_require=tests_require,
    license='SEE LICENSE IN LICENSE',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from .metaStats_models import Metrics, Trade, OpenTrade
from .domain_client import DomainClient


class MetaStatsClient:
    """metaapi.cloud MetaStats MetaTrader API client"""
    def __init__(self, domain_client: DomainClient):
        """Inits MetaTrader API client instance.

        Args:
            domain_client: HTTP client.
        """
        self._domainClient = domain_client

    async def get_metrics(self, account_id: str, include_open_positions: bool = False) -> Metrics:
        """Returns metrics of MetaApi account. This API call is billable
        https://metaapi.cloud/docs/metastats/restApi/api/calculateMetrics/

        Args:
            account_id: MetaApi account id.
            include_open_positions: Indicates whether open positions will be included in the metrics, default false.

        Returns:
            Account metrics.
        """
        def get_opts(host: str, id: str):
            return {
                'url': host + f'/users/current/accounts/{id}/metrics',
                'method': 'GET',
                'params': {
                    'includeOpenPositions': include_open_positions
                },
                'headers': {
                    'auth-token': self._domainClient.token
                }
            }

        response = await self._domainClient.request_metastats(get_opts, account_id)
        return response['metrics']

    async def get_account_trades(self, account_id: str, start_time: str, end_time: str, update_history: bool = True,
                                 limit: int = 1000, offset: int = 0) -> Trade:
        """Returns historical trades of MetaApi account.
        https://metaapi.cloud/docs/metastats/restApi/api/getHistoricalTrades/

        Args:
            account_id: MetaApi account id.
            start_time: Start of time range, inclusive.
            end_time: End of time range, exclusive.
            update_history: update historical trades before returning results. If set to true, the API call will be
            counted towards billable MetaStats API calls. If set to false, the API call is not billable.
            Default is true
            limit: Pagination limit.
            offset: Pagination offset.

        Returns:
            Account historical trades.
        """
        def get_opts(host: str, id: str):
            return {
                'url': host + f'/users/current/accounts/{id}/historical-trades/{start_time}/{end_time}',
                'method': 'GET',
                'params': {
                    'updateHistory': update_history,
                    'limit': limit,
                    'offset': offset
                },
                'headers': {
                    'auth-token': self._domainClient.token
                }
            }
        response = await self._domainClient.request_metastats(get_opts, account_id)
        return response['trades']

    async def get_account_open_trades(self, account_id: str) -> OpenTrade:
        """Returns open trades of MetaApi account. This API call is not billable.
        https://metaapi.cloud/docs/metastats/restApi/api/getOpenTrades/

        Args:
            account_id: MetaApi account id.

        Returns:
            Account historical trades.
        """
        def get_opts(host: str, id: str):
            return {
                'url': host + f'/users/current/accounts/{id}/open-trades',
                'method': 'GET',
                'headers': {
                    'auth-token': self._domainClient.token
                }
            }
        response = await self._domainClient.request_metastats(get_opts, account_id)
        return response['openTrades']
